      program adiabatic_white_dwarf
      include 'implno.dek'
      include 'const.dek'
      include 'network.dek'
      include 'vector_eos.dek'


! computes the structure of stars in hydrostatic equilibrium
! with an imposed adiabatic temperature gradient and obey a general
! stellar equation of state.


! declare
      character*80 :: string
      real*8       :: mass,temp,den,wbar,ye,xcess, &
                      presc,ener,stop_den,denc,tempc, &
                      deltam,sum,lambda,qeff

      real*8, parameter ::  con1      = 4.0d0/3.0d0 * pi, &
                            con3      = 3.0d0 /(8.0d0 * pi), &
                            third     = 1.0d0/3.0d0, &
                            twothird  = 2.0d0/3.0d0, &
                            fourthird = 4.0d0/3.0d0, & 
                            con2      = (pi/6.0d0)**third 

! declare the integration technology
      external           :: derv,rkqc
      integer            :: kount,nok,nbad,iprint,i
      integer, parameter :: xdim=600, ydim=5
      real*8             :: xrk(xdim),yrk(ydim,xdim),bc(ydim),dydx,stptry, &
                            stpmin,tol,odescal,start,sstop,dxsav

! common communication stop integrating when the pressure or temperature gets small
      real*8  :: stop_pres,stop_temp
      common  /sbyiol/ stop_pres,stop_temp

! common communication
! flag for general relativistic corrections,
! flag for isothermal evolution
! effective q-value qeff as a function of density
      logical   ::    genrel,isothermal
      real*8    ::    rho_tab(3),qeff_tab(3),background_temp
      common /grcm1/  rho_tab,qeff_tab,background_temp,genrel,isothermal


! formats
01    format(1x,i4,1pe16.8,1p10e12.4)
03    format(a)
04    format(1x,a,t10,a,t26,a,t38,a,t50,a,t62,a,t74,a,t86,a,t98,a,t110,a)


! read the helm eos tables
      call read_helm_table

! initialize the network
      call init_aprox19


! flag if general relativistic corrections are to be included
! flag if star is isothermal 

      genrel     = .false.
      isothermal = .false.


! set the composition of the object
      xmass(1:ionmax) = 1.0d-30
      xmass(ic12) = 0.5d0
      xmass(io16) = 0.5d0


! get abar and zbar and store them for either eos
      call azbar(xmass,aion,zion,wion,ionmax, &
                 ymass,abar_row(1),zbar_row(1),wbar,ye,xcess)


! fill the effective q-value qeff arrays for interpolation
! these values are from chamulak et al 2008, below equation 6
      rho_tab(1)  = 3.0e9
      qeff_tab(1) = 9.43d0 * ev2erg * 1.0e6
      rho_tab(2)  = 2.0e9
      qeff_tab(2) = 9.11d0 * ev2erg * 1.0e6
      rho_tab(3)  = 1.0e9
      qeff_tab(3) = 8.91d0 * ev2erg * 1.0e6


! get the central density
      write(6,*) 'give the central density and temperature =>'
      read(5,*) denc,tempc


! get the output file name
      write(6,*) 'give root output file name =>'
      read(5,03) string
      open(unit=3, file=string, status='unknown')


! get the equation of state of the center
      den_row(1) = denc
      temp_row(1) = tempc
      jlo_eos     = 1
      jhi_eos     = 1
      call helmeos
      presc = ptot_row(1)


! set the initial conditions:
! start is the mass point to begin the integration
! bc(1) is the gravitational radius interior to start 
! bc(2) is the pressure at the start point   
! bc(3) is the temperature at the start point
! bc(4) is the energy generation at the start point
! bc(5) is the internal energy at the start point
! these are first order expansions about zero radius/mass

      start   = 1.0d-6 * msol

! bc(1) and bc(2) first order expansions about zero radius, see for example, kippenhand and weigert equations 10.3 and 10.6
      bc(1)   = (start/(con1 * denc))**third
      bc(2)   = presc - con3 * g * (con1*denc)**fourthird * start**twothird


! for adiabatic temperature profiles a first order expansion about zero is give by kippenhand and weigert equations 10.9
      if (isothermal) then
       bc(3) = tempc
      else
       bc(3)   = exp(log(tempc) - con2 * g * nabad_row(1)/presc * denc**fourthird * start**twothird)
      end if


! lambda is the c12+c12 reaction rate
! first expression is equations 2 and 3 from woosley, wunsch & kuhlen 2004
! multiply by two because woosley et al subsume the stoichiometric factor into their rate;
! see footnote 6 in chamulak et al 2008

      lambda = 7.6d-16 * (tempc/7.0d8)**(30.0d0) * &
               1.1d3 * (denc/2.0d9)**(2.3d0) * (tempc/7.0d8)**(-7.0d0)
      lambda = 2.0d0 * lambda

! get the effective q-value ! from interpolation of the values in chamulak et al 2008, below equation 6
      if (denc .le. rho_tab(3)) then
       qeff = qeff_tab(3)
      else if (denc .ge. rho_tab(1)) then
       qeff = qeff_tab(1)
      else
       call hardwire_linear(rho_tab,qeff_tab,3,denc,qeff)
      end if


! bc(4) equation 6 of chamulak et al 2008, normalized by msol for integration scaling purposes

      bc(4) = (0.5d0*qeff*avo*ymass(ic12)**2 *denc*lambda)/msol


! the internal energy from the eos, normalized by msol for integration scaling purposes
      bc(5) = etot_row(1)/msol



! say what the initial conditions are
!111   format(1x,a,1pe14.6)
!      write(6,111) 'mass   ',start
!      write(6,111) 'radius ',bc(1)
!      write(6,111) 'pres   ',bc(2)
!      write(6,111) 'temper ',bc(3)
!      write(6,111) 'enuc   ',bc(4)
!      write(6,111) 'ener   ',bc(5)




! stop integrating conditions
! set the stopping temperature and density conditions
      if (isothermal) then
       stop_temp = tempc
      else
       stop_temp  = max(1.0d4, 1.0e-6 * tempc)
      end if
      stop_den   = max(1.0d0, 1.0e-6 * denc)

! store them and get the equivalent stopping pressure
      temp_row(1) = stop_temp
      den_row(1) = stop_den
      call helmeos
      stop_pres = ptot_row(1)

! background temperature of the white dwarf
      background_temp = 3.0d8


! control parameters
      stptry  = start
      stpmin  = start * 1.0d-8
      sstop   = 30.0 * msol
      tol     = 1.0d-6
      dxsav   = 0.0d0
      odescal = 1.0d0
      iprint  = 0


! integrate the odes in routine derv

      call cwdint(start,stptry,stpmin,sstop,bc, &
                  tol,dxsav,xdim, &
                  xrk,yrk,xdim,ydim,xdim,ydim, &
                  nok,nbad,kount,odescal,iprint, &
                  derv,rkqc)

!      write(6,*) kount


! write out the profile
      write(3,04) '  i','mass_grav','radius','pressure','density', &
                     'temp','enuc','ener','egrav','entropy'

      sum = 0.0d0
      do i=1,kount

       if (i .eq. 1) then
        deltam = xrk(i)
        den_row(1) = denc
       else
        deltam = xrk(i) - xrk(i-1)
        den_row(1) = den
       end if      


       temp_row(1) = yrk(3,i)
       ptot_row(1) = yrk(2,i)
       call invert_helm_pt_quiet
       den    = den_row(1)

       sum = sum - g*xrk(i)/yrk(1,i) * deltam

       write(3,01) i,xrk(i)/msol,yrk(1,i),yrk(2,i),den,yrk(3,i), &
                   log10(yrk(4,i)*msol),log10(yrk(5,i)*msol), &
                   -log10(abs(sum)), stot_row(1)/(kerg*avo)

      enddo


      close(unit=3)
      stop 'normal termination'
      end program adiabatic_white_dwarf









      subroutine derv(x,y,dydx)
      include 'implno.dek'
      include 'const.dek'
      include 'network.dek'
      include 'vector_eos.dek'

! this routine sets up the continuity and hydrostatic equilibrium ode's.
! x is the radial coordinate, y(1) is the gravitational mass,
! y(2) is the pressure

! declare the pass
      real*8 :: x,y(*),dydx(*)


! flag for general relativistic corrections,
! flag for isothermal evolution
! effective q-value qeff as a function of density
      logical   ::    genrel,isothermal
      real*8    ::    rho_tab(3),qeff_tab(3),background_temp
      common /grcm1/  rho_tab,qeff_tab,background_temp,genrel,isothermal


! local variables
      integer   :: i,niter
      real*8    :: radius,massg,den,ye,pres,temp,ener,cor,lambda,qeff
      real*8, parameter :: con1   = 4.0d0 * pi,  c2 = clight*clight


! map the input vector
      massg  = x
      radius = y(1)
      pres   = y(2)
      temp   = y(3)


! at this pressure and temperature, get the density
      temp_row(1) = temp
      ptot_row(1) = pres
      call invert_helm_pt_quiet
      den = den_row(1)


! here is dr/d(massg)
      dydx(1) = 1.0d0/(con1 * radius**2 * den)


! here is d(press)/d(massg)
      if (genrel) then
       cor = (1.0d0 + pres/(den*c2)) * &
             (1.0d0 + (con1*pres*radius**3)/(massg*c2)) / &
             (1.0d0 - (2.0d0*g*massg)/(radius*c2))
      else
       cor = 1.0d0
      end if

      dydx(2) = -g * massg/radius**2 * den * cor * dydx(1)


! here is the isothermal or adiabatic temperature gradient
      if (isothermal  .or. temp .le. background_temp) then
       dydx(3) = 0.0d0
      else
       dydx(3) = dydx(2) * temp/pres * nabad_row(1)
      end if


! lambda is the c12+c12 reaction rate
! first expression is equations 2 and 3 from woosley, wunsch & kuhlen 2004
! multiply by two because woosley et al subsume the stoichiometric factor into their rate;
! see footnote 6 in chamulak et al 2008

       lambda = 7.6d-16 * (temp/7.0d8)**(30.0d0) * &
               1.1d3 * (den/2.0d9)**(2.3d0) * (temp/7.0d8)**(-7.0d0)
       lambda = 2.0d0 * lambda

! get the effective q-value ! from interpolation of the values in chamulak et al 2008, below equation 6
       if (den .le. rho_tab(3)) then
        qeff = qeff_tab(3)
       else if (den .ge. rho_tab(1)) then
        qeff = qeff_tab(1)
       else
        call hardwire_linear(rho_tab,qeff_tab,3,den,qeff)
       end if

! bc(4) equation 6 of chamulak et al 2008, normalized by msol for integration scaling purposes
       dydx(4) = (0.5d0*qeff*avo*ymass(ic12)**2 *den*lambda)/msol


! internal energy from the eos, normalized by msol for integration scaling purposes
       dydx(5) = etot_row(1)/msol


      return
      end subroutine derv





      subroutine cwdint(start,stptry,stpmin,stopp,bc, &
                         eps,dxsav,kmax, &
                         xrk,yrk,xphys,yphys,xlogi,ylogi, &
                         nok,nbad,kount,odescal,iprint, &
                         derivs,steper)
      include 'implno.dek'

! generic ode driver, tuned a bit for adiabatic wd


! declare the pass
      external   :: derivs,steper
      integer    :: xphys,yphys,xlogi,ylogi,nok,nbad, &
                    kmax,kount,iprint
      real*8     :: start,stptry,stpmin,stopp,bc(yphys),eps, &
                    dxsav,xrk(xphys),yrk(yphys,xphys),odescal


! local variables
      integer            :: i,j,nstp
      integer, parameter :: nmax = 20, stpmax=10000
      real*8             :: yscal(nmax),y(nmax),dydx(nmax), &
                            x,xsav,h,hdid,hnext
      real*8, parameter  :: zero=0.0d0, one=1.0d0, tiny=1.0d-15


! stop integrating when the pressure or temperature gets small
      real*8  :: stop_pres,stop_temp
      common  /sbyiol/ stop_pres,stop_temp


! here are the format statements for printouts as we integrate
100   format(1x,i4,1p12e10.2)


! initialize
      if (ylogi .gt. yphys) stop 'ylogi > yphys in routine cwdint'
      if (yphys .gt. nmax)  stop 'yphys > nmax in routine cwdint'
      x     = start
      h     = sign(stptry,stopp-start)
      nok   = 0
      nbad  = 0
      kount = 0


! store the first step
      y(1:ylogi) = bc(1:ylogi)
      xsav = x - 2.0d0 * dxsav



! take at most stpmax steps
      do nstp=1,stpmax
       call derivs(x,y,dydx)


! scaling vector used to monitor accuracy
       do i=1,ylogi

! constant fractional accuracy
!        yscal(i) = abs(y(i)) + tiny

! const. frac. cept near zero
!        yscal(i)=abs(y(i)) + abs(h*dydx(i)) + tiny

! step size dependent accuracy
!          yscal(i) = abs(odescal * h * dydx(i)) + tiny

! for stiffs
        yscal(i) = max(odescal,y(i))

! strait scaling (decrease to get more time steps)
!        yscal(i) = odescal
       enddo



! store intermediate results
       if (kmax .gt. 0) then
        if ( (abs(dxsav) - abs(x-xsav)) .le. tiny) then
         if ( kount .lt. (kmax-1) ) then
          kount         = kount+1
          xrk(kount)    = x
          do i=1,ylogi
           yrk(i,kount) = y(i)
          enddo
          if (iprint .eq. 1) then
           write(6,100) kount,xrk(kount),(yrk(j,kount), j=1,ylogi)
          end if
          xsav=x
         end if
        end if
       end if



! if the step can overshoot the stop point or the dxsav increment then cut it
       if ((x+h-stopp)*(x+h-start) .gt. zero) h = stopp - x
       if (dxsav.ne.zero .and. h.gt.(xsav-x+dxsav)) h = xsav + dxsav-x


! do an integration step
       call steper(y,dydx,ylogi,x,h,eps,yscal,hdid,hnext,derivs)
       if (hdid.eq.h) then
        nok = nok+1
       else
        nbad = nbad+1
       end if


! this is the normal exit point, save the final step
       if (nstp .eq. stpmax .or. (x-stopp)*(stopp-start) .ge. zero &
           .or. y(2) .le. stop_pres) then
        do i=1,ylogi
         bc(i) = y(i)
        enddo
        if (kmax.ne.0) then
         kount         = kount+1
         xrk(kount)    = x
         do i=1,ylogi
          yrk(i,kount) = y(i)
         enddo
         if (iprint .eq. 1) then
           write(6,100) kount,xrk(kount),(yrk(j,kount), j=1,ylogi)
         end if
        end if
        return
       end if


! set the step size for the next iteration; stay above stpmin
!       write(6,'(1x,1p4e14.6)') h, hdid, hnext, stpmin

       h=hnext

       if (abs(hnext).lt.stpmin) return
!       if (abs(hnext).lt.stpmin) stop 'hnext < stpmin in cwdint'


! back for another iteration or death
      enddo
      write(6,*) '> than stpmax steps required in cwdint'
      return
      end subroutine cwdint




      subroutine rkqc(y,dydx,n,x,htry,eps,yscal,hdid,hnext,derivs)
      include 'implno.dek'

! fifth order, step doubling, runge-kutta ode integrator with monitering of
! local truncation errors. input are the vector y of length n, which has a
! known the derivative dydx at the point x, the step size to be attempted
! htry, the required accuracy eps, and the vector yscal against which the
! error is to be scaled. on output, y and x are replaced by their new values,
! hdid is the step size that was actually accomplished, and hnext is the
! estimated next step size. derivs is a user supplied routine that computes
! the right hand side of the first order system of ode's. plug into odeint.

! declare the pass
      external    :: derivs
      integer     :: n
      real*8      :: x,htry,eps,hdid,hnext,y(n),dydx(n),yscal(n)


! local variables
      integer             :: i
      integer, parameter  :: nmax = 2000
      real*8              :: ytemp(nmax),ysav(nmax),dysav(nmax),xsav,h,hh,errmax
      real*8, parameter   :: fcor=1.0d0/15.0d0, pgrow = -0.2d0, &
                             pshrnk = -0.25d0,  safety=0.9d0,  errcon=6.0e-4

! note errcon = (4/safety)**(1/pgrow)
! nmax is the maximum number of differential equations


! save the initial values
      h      = htry
      xsav   =  x
      do i=1,n
       ysav(i)  = y(i)
       dysav(i) = dydx(i)
      enddo

! take two half steps
1     hh = 0.5d0*h
      call rk4(ysav,dysav,n,xsav,hh,ytemp,derivs)
      x  = xsav + hh
      call derivs(x,ytemp,dydx)
      call rk4(ytemp,dydx,n,x,hh,y,derivs)
      x  = xsav + h
      if (x .eq. xsav) stop 'stepsize not significant in rkqc'

! now take the large step
      call rk4(ysav,dysav,n,xsav,h,ytemp,derivs)

! ytemp is the error estimate
      errmax = 0.0
      do i=1,n
       ytemp(i) = y(i) - ytemp(i)
       errmax   = max(errmax,abs(ytemp(i)/yscal(i)))
      enddo
      errmax     = errmax/eps

! truncation error too big, reduce the step size and try again
      if (errmax .gt. 1.0) then
       h = safety * h * (errmax**pshrnk)
       go to  1

! truncation within limits, compute the size of the next step
      else
       hdid = h
       if (errmax.gt.errcon) then
        hnext = safety * h * (errmax**pgrow)
       else
        hnext = 4.0d0 * h
       end if
      end if

! mop up the fifth order truncation error
      do i=1,n
       y(i) = y(i) + ytemp(i)*fcor
      enddo
      return
      end subroutine rkqc




       subroutine rk4(y,dydx,n,x,h,yout,derivs)
       include 'implno.dek'

! given values for the variables y(1:n) and their derivatives dydx(1:n) known
! at x, use the fourth order runge-kutta method to advance the solution over
! an interval h and return the incremented variables in yout(1:n) (which need
! not be a distinct array from y). one supplies the routine derivs which
! evaluates the right hand side of the ode's.

! declare the pass
       external   :: derivs
       integer    :: n
       real*8     :: x,h,y(n),dydx(n),yout(n)


! local variables
       integer            :: i
       integer, parameter :: nmax = 2000
       real*8             :: yt(nmax),dyt(nmax),dym(nmax),hh,h6,xh

! initialize the step sizes and weightings
       hh = h*0.5d0
       h6 = h/6.0d0
       xh = x + hh

! the first step
       do i=1,n
        yt(i) = y(i) + hh*dydx(i)
       enddo

! the second step
       call derivs(xh,yt,dyt)
       do i=1,n
        yt(i) = y(i) + hh*dyt(i)
       enddo

! the third step
       call derivs(xh,yt,dym)
       do i=1,n
        yt(i)  = y(i) + h*dym(i)
        dym(i) = dyt(i) + dym(i)
       enddo

! the fourth step and accumulate the increments with the proper weights
       call derivs(x+h,yt,dyt)
       do i=1,n
        yout(i) = y(i) + h6*(dydx(i) +dyt(i) + 2.0d0*dym(i))
       enddo
       return
       end subroutine rk4



      subroutine hardwire_linear(xtab,ytab,n,xin,yout)
      implicit none
      save

! implements a hardwired linear interpolation

! declare the pass
      integer ::       n
      real*8  :: xtab(n),ytab(n),xin,yout

! local variables
      real*8 :: x1,x2,denom,alfa,beta

      x1    = xtab(1)
      x2    = xtab(2)
      denom = 1.0d0/(x1-x2)
      alfa  = (xin - x2)*denom
      beta  = (x1 - xin)*denom
      yout  = alfa*ytab(1) + beta*ytab(2)

      return
      end



      subroutine azbar(xmass,aion,zion,wion,ionmax, &
                       ymass,abar,zbar,wbar,ye,nxcess)
      include 'implno.dek'

! this routine calculates composition variables

! input:
! mass fractions               = xmass(1:ionmax)  dimensionless
! number of nucleons           = aion(1:ionmax)   dimensionless
! charge of nucleus            = zion(1:ionmax)   dimensionless
! atomic weight or molar mass  = wion(1:ionmax)   dimensionless
! number of isotopes           = ionmax
!
! output:
! molar abundances        = ymass(1:ionmax)   dimensionless
! mean number of nucleons = abar              dimensionless
! mean nucleon charge     = zbar              dimensionless
! mean weight             = wbar              dimensionless
! electron fraction       = ye                dimensionaless
! neutron excess          = xcess             dimensionless


! declare the pass
      integer    ::  ionmax
      real*8     ::  xmass(ionmax),aion(ionmax),zion(ionmax), &
                     wion(ionmax),ymass(ionmax),abar,zbar,wbar, &
                     ye,nxcess


! local variables
      real*8 :: sum1

! molar abundances
      ymass(1:ionmax) = xmass(1:ionmax)/aion(1:ionmax)

! mean numner of nucleons
      abar  = 1.0d0/sum(ymass(1:ionmax))
      wbar  = abar

! mean charge
      ye  = sum(zion(1:ionmax)*ymass(1:ionmax))
      zbar  = abar * ye

! neutron excess
      nxcess = 1.0d0 - 2.0d0 * ye

      return
      end subroutine azbar



      subroutine init_aprox19
      include 'implno.dek'
      include 'const.dek'
      include 'network.dek'
!
! this routine initializes stuff for the aprox19 network
!
! declare
      integer           :: i
      real*8, parameter :: mev2erg = ev2erg*1.0d6, mev2gr  = mev2erg/clight**2


! zero all the isotope and rate pointers
      call init_isotope_rate_pointers


! set the size of the network and the number of rates
      idnet   = idaprox19
      ionmax  = 19
      iener   = ionmax + 1
      itemp   = ionmax + 2
      iden    = ionmax + 3
      ivelx   = ionmax + 4
      iposx   = ionmax + 5
      neqs    = iposx
      nrat    = 101
      netname = 'aprox19'


! set the id numbers of the elements
      ih1   = 1
      ihe3  = 2
      ihe4  = 3
      ic12  = 4
      in14  = 5
      io16  = 6
      ine20 = 7
      img24 = 8
      isi28 = 9
      is32  = 10
      iar36 = 11
      ica40 = 12
      iti44 = 13
      icr48 = 14
      ife52 = 15
      ife54 = 16
      ini56 = 17
      ineut = 18
      iprot = 19


! set the names of the elements
      ionam(ih1)   = 'h1  '
      ionam(ihe3)  = 'he3 '
      ionam(ihe4)  = 'he4 '
      ionam(ic12)  = 'c12 '
      ionam(in14)  = 'n14 '
      ionam(io16)  = 'o16 '
      ionam(ine20) = 'ne20'
      ionam(img24) = 'mg24'
      ionam(isi28) = 'si28'
      ionam(is32)  = 's32 '
      ionam(iar36) = 'ar36'
      ionam(ica40) = 'ca40'
      ionam(iti44) = 'ti44'
      ionam(icr48) = 'cr48'
      ionam(ife52) = 'fe52'
      ionam(ife54) = 'fe54'
      ionam(ini56) = 'ni56'
      ionam(ineut) = 'neut'
      ionam(iprot) = 'prot'

      ionam(iener) = 'ener '
      ionam(itemp) = 'temp '
      ionam(iden)  = 'den  '
      ionam(ivelx) = 'velx '
      ionam(iposx) = 'posx '


! set the number of nucleons in the element
      aion(ih1)   = 1.0d0
      aion(ihe3)  = 3.0d0
      aion(ihe4)  = 4.0d0
      aion(ic12)  = 12.0d0
      aion(in14)  = 14.0d0
      aion(io16)  = 16.0d0
      aion(ine20) = 20.0d0
      aion(img24) = 24.0d0
      aion(isi28) = 28.0d0
      aion(is32)  = 32.0d0
      aion(iar36) = 36.0d0
      aion(ica40) = 40.0d0
      aion(iti44) = 44.0d0
      aion(icr48) = 48.0d0
      aion(ife52) = 52.0d0
      aion(ife54) = 54.0d0
      aion(ini56) = 56.0d0
      aion(ineut) = 1.0d0
      aion(iprot) = 1.0d0

! set the number of protons in the element
      zion(ih1)   = 1.0d0
      zion(ihe3)  = 2.0d0
      zion(ihe4)  = 2.0d0
      zion(ic12)  = 6.0d0
      zion(in14)  = 7.0d0
      zion(io16)  = 8.0d0
      zion(ine20) = 10.0d0
      zion(img24) = 12.0d0
      zion(isi28) = 14.0d0
      zion(is32)  = 16.0d0
      zion(iar36) = 18.0d0
      zion(ica40) = 20.0d0
      zion(iti44) = 22.0d0
      zion(icr48) = 24.0d0
      zion(ife52) = 26.0d0
      zion(ife54) = 26.0d0
      zion(ini56) = 28.0d0
      zion(ineut) = 0.0d0
      zion(iprot) = 1.0d0


! set the binding energy of the element
      bion(ih1)   = 0.0d0
      bion(ihe3)  = 7.71819d0
      bion(ihe4)  =  28.29603d0
      bion(ic12)  =  92.16294d0
      bion(in14)  = 104.65998d0
      bion(io16)  = 127.62093d0
      bion(ine20) = 160.64788d0
      bion(img24) = 198.25790d0
      bion(isi28) = 236.53790d0
      bion(is32)  = 271.78250d0
      bion(iar36) = 306.72020d0
      bion(ica40) = 342.05680d0
      bion(iti44) = 375.47720d0
      bion(icr48) = 411.46900d0
      bion(ife52) = 447.70800d0
      bion(ife54) = 471.7696d0
      bion(ini56) = 484.00300d0
      bion(ineut) = 0.0d0
      bion(iprot) = 0.0d0


! set the number of neutrons and mass
      do i=1,ionmax
       nion(i) = aion(i) - zion(i)
      enddo

! mass of each isotope
      do i = 1,ionmax
       mion(i) = nion(i)*mn + zion(i)*mp - bion(i)*mev2gr
      enddo

! molar mass
      do i = 1,ionmax
       wion(i) = avo * mion(i)
      enddo

! a common approximation
      do i = 1,ionmax
       wion(i) = aion(i)
      enddo



! set the partition functions - statistical weights, ground-state only here
      do i=1,ionmax
       wpart(i) = 1.0d0
      enddo
      wpart(ih1)   = 1.0d0
      wpart(ineut) = 1.0d0
      wpart(iprot) = 1.0d0


! set the id numbers of the reaction rates
      ir3a   = 1
      irg3a  = 2
      ircag  = 3
      ir1212 = 4
      ir1216 = 5
      ir1616 = 6
      iroga  = 7
      iroag  = 8
      irnega = 9
      irneag = 10
      irmgga = 11
      irmgag = 12
      irsiga = 13
      irmgap = 14
      iralpa = 15
      iralpg = 16
      irsigp = 17
      irsiag = 18
      irsga  = 19
      irsiap = 20
      irppa  = 21
      irppg  = 22
      irsgp  = 23
      irsag  = 24
      irarga = 25
      irsap  = 26
      irclpa = 27
      irclpg = 28
      irargp = 29
      irarag = 30
      ircaga = 31
      irarap = 32
      irkpa  = 33
      irkpg  = 34
      ircagp = 35
      ircaag = 36
      irtiga = 37
      ircaap = 38
      irscpa = 39
      irscpg = 40
      irtigp = 41
      irtiag = 42
      ircrga = 43
      irtiap = 44
      irvpa  = 45
      irvpg  = 46
      ircrgp = 47
      ircrag = 48
      irfega = 49
      ircrap = 50
      irmnpa = 51
      irmnpg = 52
      irfegp = 53
      irfeag = 54
      irniga = 55
      irfeap = 56
      ircopa = 57
      ircopg = 58
      irnigp = 59

! for fe54 photodisintegration
      ir52ng = 60
      ir53gn = 61
      ir53ng = 62
      ir54gn = 63
      irfepg = 64
      ircogp = 65

! for he4 photodisintegration
      irheng = 66
      irhegn = 67
      irhng  = 68
      irdgn  = 69
      irdpg  = 70
      irhegp = 71

! weak reactions
      irpen   = 72
      irnep   = 73
      irn56ec = 74

! ppchain
      irpp    = 75
      ir33    = 76
      irhe3ag = 77

! cno cycles
      ircpg  = 78
      irnpg  = 79
      ifa    = 80
      ifg    = 81
      iropg  = 82
      irnag  = 83

! the dummy links
      irr1   = 84
      irs1   = 85
      irt1   = 86
      iru1   = 87
      irv1   = 88
      irw1   = 89
      irx1   = 90
      iry1   = 91

      ir1f54 = 92
      ir2f54 = 93
      ir3f54 = 94
      ir4f54 = 95
      ir5f54 = 96
      ir6f54 = 97
      ir7f54 = 98
      ir8f54 = 99

      iralf1 = 100
      iralf2 = 101



! set the names of the reaction rates
      ratnam(ir3a)   = 'r3a  '
      ratnam(irg3a)  = 'rg3a '
      ratnam(ircag)  = 'rcag '
      ratnam(ir1212) = 'r1212'
      ratnam(ir1216) = 'r1216'
      ratnam(ir1616) = 'r1616'
      ratnam(iroga)  = 'roga '
      ratnam(iroag)  = 'roag '
      ratnam(irnega) = 'rnega'
      ratnam(irneag) = 'rneag'
      ratnam(irmgga) = 'rmgga'
      ratnam(irmgag) = 'rmgag'
      ratnam(irsiga) = 'rsiga'
      ratnam(irmgap) = 'rmgap'
      ratnam(iralpa) = 'ralpa'
      ratnam(iralpg) = 'ralpg'
      ratnam(irsigp) = 'rsigp'
      ratnam(irsiag) = 'rsiag'
      ratnam(irsga)  = 'rsga '
      ratnam(irsiap) = 'rsiap'
      ratnam(irppa)  = 'rppa '
      ratnam(irppg)  = 'rppg '
      ratnam(irsgp)  = 'rsgp '
      ratnam(irsag)  = 'rsag '
      ratnam(irarga) = 'rarga'
      ratnam(irsap)  = 'rsap '
      ratnam(irclpa) = 'rclpa'
      ratnam(irclpg) = 'rclpg'
      ratnam(irargp) = 'rargp'
      ratnam(irarag) = 'rarag'
      ratnam(ircaga) = 'rcaga'
      ratnam(irarap) = 'rarap'
      ratnam(irkpa)  = 'rkpa '
      ratnam(irkpg)  = 'rkpg '
      ratnam(ircagp) = 'rcagp'
      ratnam(ircaag) = 'rcaag'
      ratnam(irtiga) = 'rtiga'
      ratnam(ircaap) = 'rcaap'
      ratnam(irscpa) = 'rscpa'
      ratnam(irscpg) = 'rscpg'
      ratnam(irtigp) = 'rtigp'
      ratnam(irtiag) = 'rtiag'
      ratnam(ircrga) = 'rcrga'
      ratnam(irtiap) = 'rtiap'
      ratnam(irvpa)  = 'rvpa '
      ratnam(irvpg)  = 'rvpg '
      ratnam(ircrgp) = 'rcrgp'
      ratnam(ircrag) = 'rcrag'
      ratnam(irfega) = 'rfega'
      ratnam(ircrap) = 'rcrap'
      ratnam(irmnpa) = 'rmnpa'
      ratnam(irmnpg) = 'rmnpg'
      ratnam(irfegp) = 'rfegp'
      ratnam(irfeag) = 'rfeag'
      ratnam(irniga) = 'rniga'
      ratnam(irfeap) = 'rfeap'
      ratnam(ircopa) = 'rcopa'
      ratnam(ircopg) = 'rcopg'
      ratnam(irnigp) = 'rnigp'

! for fe54 photodisintegration
      ratnam(ir52ng) = 'r52ng'
      ratnam(ir53gn) = 'r53gn'
      ratnam(ir53ng) = 'r53ng'
      ratnam(ir54gn) = 'r54gn'
      ratnam(irfepg) = 'rfepg'
      ratnam(ircogp) = 'rcogp'

! for he4 photodisintegration
      ratnam(irheng)  = 'rheng'
      ratnam(irhegn)  = 'rhegn'
      ratnam(irhng)   = 'rhng '
      ratnam(irdgn)   = 'rdgn '
      ratnam(irdpg)   = 'rdpg '
      ratnam(irhegp)  = 'rhegp'

! for weak reactions
      ratnam(irpen)   = 'rpen '
      ratnam(irnep)   = 'rnep '
      ratnam(irn56ec) = 'r56ec'

! ppchain
      ratnam(irpp)    = 'rpp  '
      ratnam(ir33)    = 'r33  '
      ratnam(irhe3ag) = 'rhe3ag'

! cno cycles
      ratnam(ircpg)   = 'rcpg '
      ratnam(irnpg)   = 'rnpg '
      ratnam(iropg)   = 'ropg '
      ratnam(ifa)     = 'rfa  '
      ratnam(ifg)     = 'rfg  '
      ratnam(irnag)   = 'rnag '


! the dummy links
      ratnam(irr1)   = 'r1   '
      ratnam(irs1)   = 's1   '
      ratnam(irt1)   = 't1   '
      ratnam(iru1)   = 'u1   '
      ratnam(irv1)   = 'v1   '
      ratnam(irw1)   = 'w1   '
      ratnam(irx1)   = 'x1   '
      ratnam(iry1)   = 'y1   '


      ratnam(ir1f54) = 'r1f54'
      ratnam(ir2f54) = 'r2f54'
      ratnam(ir3f54) = 'r3f54'
      ratnam(ir4f54) = 'r4f54'
      ratnam(ir5f54) = 'r5f54'
      ratnam(ir6f54) = 'r6f54'
      ratnam(ir7f54) = 'r7f54'
      ratnam(ir8f54) = 'r8f54'

      ratnam(iralf1) = 'ralf1'
      ratnam(iralf2) = 'ralf2'


      return
      end subroutine init_aprox19



      subroutine init_isotope_rate_pointers
      include 'implno.dek'
      include 'network.dek'

! initializes all the integer pointers for the isotopes and rates


! zero all the isotope pointers
      ih1=0   ; iprot=0 ; ineut=0 ; ihe4=0  ; ih2=0   ; ih3=0   ; ihe3=0  ; ili6=0  ; ili7=0  ; ili8=0
      ibe7=0  ; ibe9=0  ; ib8=0   ; ib9=0   ; ib10=0  ; ib11=0  ; ic11=0  ; ic12=0  ; ic13=0  ; ic14=0
      in12=0  ; in13=0  ; in14=0  ; in15=0  ; io14=0  ; io15=0  ; io16=0  ; io17=0  ; io18=0  ; if17=0
      if18=0  ; if19=0  ; ine18=0 ; ine19=0 ; ine20=0 ; ine21=0 ; ine22=0 ; ine23=0 ; ina20=0 ; ina21=0
      ina22=0 ; ina23=0 ; img22=0 ; img23=0 ; img24=0 ; img25=0 ; img26=0 ; ial25=0 ; ial26=0 ; ial27=0
      ial28=0 ; isi26=0 ; isi27=0 ; isi28=0 ; isi29=0 ; isi30=0 ; ip28=0  ; ip29=0  ; ip30=0  ; ip31=0
      is30=0  ; is31=0  ; is32=0  ; is33=0  ; is34=0  ; icl35=0 ; icl37=0 ; iar36=0 ; iar37=0 ; iar38=0
      ik39=0  ; ica40=0 ; isc43=0 ; isc45=0 ; iti44=0 ; iti46=0 ; iti48=0 ; iti50=0 ; iv46=0  ; iv47=0
      iv48=0  ; iv51=0  ; icr47=0 ; icr48=0 ; icr49=0 ; icr50=0 ; icr51=0 ; icr52=0 ; icr53=0 ; icr54=0
      icr56=0 ; imn50=0 ; imn51=0 ; imn52=0 ; imn55=0 ; ife52=0 ; ife54=0 ; ife55=0 ; ife56=0 ; ife57=0
      ife58=0 ; ife59=0 ; ico54=0 ; ico55=0 ; ico56=0 ; ico59=0 ; ini56=0 ; ini57=0 ; ini58=0 ; ini59=0
      ini64=0 ; ini66=0 ; icu63=0 ; izn60=0 ; izn64=0


! zero all the rate pointers

      ir3a=0   ;  irg3a=0  ;  ircag=0  ;  ir1212=0 ;  ir1216=0 ;  iroga=0 
      iroag=0  ;  irnega=0 ;  irneag=0 ;  irmgga=0 ;  irmgag=0 ;  irsiga=0   
      irmgap=0 ;  iralpa=0 ;  iralpg=0 ;  irsigp=0 ;  irsiag=0 ;  irsga=0    
      irsiap=0 ;  irppa=0  ;  irppg=0  ;  irsgp=0  ;  irsag=0  ;  irarga=0   
      irsap=0  ;  irclpa=0 ;  irclpg=0 ;  irargp=0 ;  irarag=0 ;  ircaga=0   
      irarap=0 ;  irkpa=0  ;  irkpg=0  ;  ircagp=0 ;  ircaag=0 ;  irtiga=0   
      ircaap=0 ;  irscpa=0 ;  irscpg=0 ;  irtigp=0 ;  irtiag=0 ;  ircrga=0   
      irtiap=0 ;  irvpa=0  ;  irvpg=0  ;  ircrgp=0 ;  ircrag=0 ;  irfega=0   
      ircrap=0 ;  irmnpa=0 ;  irmnpg=0 ;  irfegp=0 ;  irfeag=0 ;  irniga=0   
      ir1616=0 ;  iraan=0

      irpp=0   ;  ir33=0   ;  ircpg=0  ;  irnpg=0  ;  iropg=0   ;  irnag=0 
      irfeap=0 ;  ircopa=0 ;  ircopg=0 ;  irnigp=0 ;  irfepg=0  ;  ircogp=0 
      ir52ng=0 ;  ir53gn=0 ;  ir53ng=0 ;  ir54gn=0 ;  irheng=0  ;  irhegn=0   
      irhng=0  ;  irhegp=0 ;  irpen=0  ;  irnep=0  ;  irn56ec=0 ;  ifa=0      
      ifg=0    

      ir54ng=0      ; ir55gn=0       ; ir55ng=0      ; ir56gn=0    
      irfe54ap=0    ; irco57pa=0     ; irfe56pg=0    ; irco57gp=0 
      irfe56_aux1=0 ;  irfe56_aux2=0 ; irfe56_aux3=0 ; irfe56_aux4=0

      irbepg=0 ;  irb8ep=0 ;  irbeec=0 ;  irpep=0 ;  irhep=0

      irc12pg=0  ;  irn13gp=0  ;  irn13enu=0  ;  irc13pg=0   ; irn14gp=0 
      irn14pg=0  ;  iro15gp=0  ;  iro15enu=0  ;  irn15pa=0   ; irc12ap=0    
      irn15pg=0  ;  iro16gp=0  ;  iro16pg=0   ;  irf17gp=0   ; irf17enu=0   
      iro17pa=0  ;  irn14ap=0  ;  iro17pg=0   ;  irf18gp=0   ; irf18enu=0   
      iro18pa=0  ;  irn15ap=0  ;  iro18pg=0   ;  irf19gp=0   ; irf19pa=0    
      iro16ap=0  ;  irn13pg=0  ;  iro14gp=0   ;  iro14enu=0  ; iro14ap=0    
      irf17pa=0  ;  irf17pg=0  ;  irne18gp=0  ;  irne18enu=0 ; irf18pa=0    
      iro15ap=0  ;  irne18ap=0 ;  iro15ag=0   ;  irne19ga=0  ; irne19pg=0   
      irna20gp=0 ;  irsi26ap=0 ;  irne19enu=0 ;  irne20pg=0  ; irna21gp=0

      irweakn=0 ;     irweakp=0 ;     irpng=0 ;       irdgn=0 ;      irdpn=0 
      ir2pnp=0 ;      irdng=0 ;       irtgn=0 ;       irddp=0 ;      irtpd=0       
      irdpg=0 ;       irhe3gp=0 ;     irddn=0 ;       irhe3nd=0 ;    irtpn=0       
      irhe3np=0 ;     irddg=0 ;       irhe4gd=0 ;     irtpg=0 ;      irhe4gp=0 
      irtdn=0 ;       irhe4nd=0 ;     irtt2n=0 ;      irhe42nt=0 ;   irhe3ng=0     
      irhe4gn=0 ;     irhe3dp=0 ;     irhe4pd=0 ;     irhe3td=0 ;    irhe4dt=0     
      irhe3tnp=0 ;    irhe4pnt=0 ;    ir33inv=0 ;     irhe4npg=0 ;   irli6gpn=0    
      irdag=0 ;       irli6ga=0 ;     irtan=0 ;       irli6na=0 ;    irli6pa=0     
      irhe3ap=0 ;     irtag=0 ;       irli7ga=0 ;     irli7pa=0 ;    ir2he4ga=0 
      irli7dn=0 ;     ir2and=0 ;      irli7t2n=0 ;    ir2a2nt=0 ;    irli7he3np=0 
      ir2apnhe3=0 ;   irhe3ag=0 ;     irbe7ga=0 ;     irli6pg=0 ;    irbe7gp=0   
      irli7pn=0 ;     irbe7np=0 ;     irbe7dp=0 ;     ir2apd=0 ;     irbe7tnp=0    
      ir2apnt=0 ;     irbe7he32p=0 ;  ir2a2phe3=0 ;   irbe9pa=0 ;    irli6ap=0     
      irbe7pg=0 ;     irb8gp=0 ;      irli6ag=0 ;     irb10ga=0 ;    irli7an=0     
      irb10na=0 ;     irbe9pg=0 ;     irb10gp=0 ;     irb10pa=0 ;    irbe7ap=0     
      irli7ag=0 ;     irb11ga=0 ;     irb11pa=0 ;     ir2he4ap=0 ;   irbe7ag=0     
      irc11ga=0 ;     irb11pn=0 ;     irc11np=0 ;     irb8ap=0 ;     irc11pa=0      
      irb10pg=0 ;     irc11gp=0 ;     irc11na=0 ;     ir3an=0 ;      irbe9an=0     
      irc12na=0 ;     irb11pg=0 ;     irc12gp=0 ;     irb11ap=0 ;    irc14pa=0     
      irli6ng=0 ;     irli7gn=0


      irh3em=0 ;    irc11em=0 ;   iraanbe7=0 ;  irdap=0 ;     irtap=0 ;   irb10ng=0 
      irbe9dn=0 ;   irb10dp=0 ;   irb8npa=0 ;   irbe9pda=0 ;  ir2adp=0 ;  irb11p2a=0 
      irc11n2a=0


      irh3pg=0 ;    irli7pag=0 ;  irgaan=0 ;    irbe9pd=0 ;   ir3ap=0 ;    ir1212p=0 ;   irna23pc=0 
      ir1212n=0 ;   irmg23nc=0 ;  ir1212a=0 ;   irne20ac=0 ;  ir1616n=0 ;  irs31no=0 ;   ir1616p=0    
      irp31po=0 ;   ir1616a=0 ;   irsi28ao=0 ;  ir1616d=0 ;   irp30do=0 ;  ir1216n=0 ;   irsi27no=0   
      ir1216p=0 ;   iral27po=0 ;  ir1216a=0 ;   irmg24ao=0 ;  irli7ng=0 ;  irli8gn=0 ;   irbe9pn=0    
      irb9np=0 ;    irb10an=0 ;   irn13na=0 ;   irb11an=0 ;   irn14na=0 ;  irn14pa=0 ;   irc11ap=0    
      irc11pg=0 ;   irn12gp=0 ;   irc12an=0 ;   iro15na=0 ;   irc13an=0 ;  iro16na=0 ;   irc13pn=0    
      irn13np=0 ;   irc14pg=0 ;   irn15gp=0 ;   irc14pn=0 ;   irn14np=0 ;  irc14ag=0 ;   iro18ga=0    
      iro16pa=0 ;   irn13ap=0 ;   irn14pn=0 ;   iro14np=0 ;   irn14an=0 ;  irf17na=0 ;   irn14ag=0    
      irf18ga=0 ;   irn15pn=0 ;   iro15np=0 ;   irn15an=0 ;   irf18na=0 ;  irn15ag=0 ;   irf19ga=0    
      iro14ag=0 ;   irne18ga=0 ;  iro17ag=0 ;   irne21ga=0 ;  iro17an=0 ;  irne20na=0 ;  iro18ag=0    
      irne22ga=0 ;  iro18an=0 ;   irne21na=0

      irne20pa=0 ;  irf17ap=0 ;   irf18pg=0 ;   irne19gp=0 ;  irf19pg=0 ;   irne20gp=0 ;  irf19pn=0    
      irne19np=0 ;  irf19ap=0 ;   irne22pa=0 ;  irna22na=0 ;  irf19an=0 ;   irna23pa=0 ;  irne20ap=0    
      irne20ng=0 ;  irne21gn=0 ;  irne21pg=0 ;  irna22gp=0 ;  irne21ag=0 ;  irmg25ga=0 ;  irne21an=0   
      irmg24na=0 ;  irne22pg=0 ;  irna23gp=0 ;  irne22ag=0 ;  irmg26ga=0 ;  irna22np=0 ;  irne22pn=0   
      irne22an=0 ;  irmg25na=0 ;  irna21pg=0 ;  irmg22gp=0 ;  irmg24pa=0 ;  irna21ap=0 ;  irna22pg=0   
      irmg23gp=0 ;  irna23pg=0 ;  irmg24gp=0 ;  irna23pn=0 ;  irmg23np=0 ;  irmg24pg=0 ;  iral25gp=0   
      iral27pa=0 ;  irmg24ap=0 ;  irmg25pg=0 ;  iral26gp=0 ;  irmg25ap=0 ;  iral28pa=0 ;  irmg25ag=0   
      irsi29ga=0 ;  irmg25an=0 ;  irsi28na=0 ;  irmg26pg=0 ;  iral27gp=0 ;  irmg26ag=0 ;  irsi30ga=0   
      irmg26an=0 ;  irsi29na=0 ;  iral25pg=0 ;  irsi26gp=0 ;  iral26pg=0 ;  irsi27gp=0 ;  iral27pg=0   
      irsi28gp=0 ;  iral27an=0 ;  irp30na=0 ;   irsi27pg=0 ;  irp28gp=0 ;   irsi28pg=0 ;  irp29gp=0    
      irsi29pg=0 ;  irp30gp=0 ;   irsi30pg=0 ;  irp31gp=0 ;   irr1=0 ;      irs1=0 ;      irt1=0       
      iru1=0 ;      irv1=0 ;      irw1=0 ;      irx1=0 ;      iry1=0 ;      ir1f54=0 ;    ir2f54=0     
      ir3f54=0 ;    ir4f54=0 ;    ir5f54=0 ;    ir6f54=0 ;    ir7f54=0 ;    ir8f54=0 ;    iralf1=0     
      iralf2=0 ;    irsi2ni=0 ;   irni2si=0 ;   iralam1=0 ;   irdelta1=0 ;  iralam2=0 ;   irdelta2=0 
      irdebug=0

      return 
      end subroutine init_isotope_rate_pointers


      subroutine read_helm_table
      include 'implno.dek'
      include 'helm_table_storage.dek'

! this routine reads the helmholtz eos file, and
! must be called once before the helmeos routine is invoked.

! declare local variables
      integer          i,j
      double precision tsav,dsav,dth,dt2,dti,dt2i,dt3i, &
                       dd,dd2,ddi,dd2i,dd3i


! open the file (use softlinks to input the desired table)

       open(unit=19,file='helm_table.dat',status='old')


! for standard table limits
       tlo   = 3.0d0
       thi   = 13.0d0
       tstp  = (thi - tlo)/float(jmax-1)
       tstpi = 1.0d0/tstp
       dlo   = -12.0d0
       dhi   = 15.0d0
       dstp  = (dhi - dlo)/float(imax-1)
       dstpi = 1.0d0/dstp

! read the helmholtz free energy and its derivatives
       do j=1,jmax
        tsav = tlo + (j-1)*tstp
        t(j) = 10.0d0**(tsav)
        do i=1,imax
         dsav = dlo + (i-1)*dstp
         d(i) = 10.0d0**(dsav)
         read(19,*) f(i,j),fd(i,j),ft(i,j),fdd(i,j),ftt(i,j),fdt(i,j), &
                  fddt(i,j),fdtt(i,j),fddtt(i,j)
        enddo
       enddo


! read the pressure derivative with density table
       do j=1,jmax
        do i=1,imax
         read(19,*) dpdf(i,j),dpdfd(i,j),dpdft(i,j),dpdfdt(i,j)
        enddo
       enddo

! read the electron chemical potential table
       do j=1,jmax
        do i=1,imax
         read(19,*) ef(i,j),efd(i,j),eft(i,j),efdt(i,j)
        enddo
       enddo

! read the number density table
       do j=1,jmax
        do i=1,imax
         read(19,*) xf(i,j),xfd(i,j),xft(i,j),xfdt(i,j)
        enddo
       enddo

! close the file and write a summary message
      close(unit=19)


! construct the temperature and density deltas and their inverses
       do j=1,jmax-1
        dth          = t(j+1) - t(j)
        dt2         = dth * dth
        dti         = 1.0d0/dth
        dt2i        = 1.0d0/dt2
        dt3i        = dt2i*dti
        dt_sav(j)   = dth
        dt2_sav(j)  = dt2
        dti_sav(j)  = dti
        dt2i_sav(j) = dt2i
        dt3i_sav(j) = dt3i
       end do
       do i=1,imax-1
        dd          = d(i+1) - d(i)
        dd2         = dd * dd
        ddi         = 1.0d0/dd
        dd2i        = 1.0d0/dd2
        dd3i        = dd2i*ddi
        dd_sav(i)   = dd
        dd2_sav(i)  = dd2
        ddi_sav(i)  = ddi
        dd2i_sav(i) = dd2i
        dd3i_sav(i) = dd3i
       enddo



!      write(6,*)
!      write(6,*) 'finished reading eos table'
!      write(6,04) 'imax=',imax,' jmax=',jmax
!04    format(1x,4(a,i4))
!      write(6,03) 'temp(1)   =',t(1),' temp(jmax)   =',t(jmax)
!      write(6,03) 'ye*den(1) =',d(1),' ye*den(imax) =',d(imax)
!03    format(1x,4(a,1pe11.3))
!      write(6,*)

      return
      end







      subroutine helmeos
      include 'implno.dek'
      include 'const.dek'
      include 'vector_eos.dek'
      include 'helm_table_storage.dek'


! given a temperature temp [K], density den [g/cm**3], and a composition
! characterized by abar and zbar, this routine returns most of the other
! thermodynamic quantities. of prime interest is the pressure [erg/cm**3],
! specific thermal energy [erg/gr], the entropy [erg/g/K], along with
! their derivatives with respect to temperature, density, abar, and zbar.
! other quantites such the normalized chemical potential eta (plus its
! derivatives), number density of electrons and positron pair (along
! with their derivatives), adiabatic indices, specific heats, and
! relativistically correct sound speed are also returned.
!
! this routine assumes planckian photons, an ideal gas of ions,
! and an electron-positron gas with an arbitrary degree of relativity
! and degeneracy. interpolation in a table of the helmholtz free energy
! is used to return the electron-positron thermodynamic quantities.
! all other derivatives are analytic.
!
! references: cox & giuli chapter 24 ; timmes & swesty apj 1999


! declare
      integer          i,j
      double precision temp,den,abar,zbar,ytot1,ye, &
                       x,y,zz,zzi,deni,tempi,xni,dxnidd,dxnida, &
                       dpepdt,dpepdd,deepdt,deepdd,dsepdd,dsepdt, &
                       dpraddd,dpraddt,deraddd,deraddt,dpiondd,dpiondt, &
                       deiondd,deiondt,dsraddd,dsraddt,dsiondd,dsiondt, &
                       dse,dpe,dsp,kt,ktinv,prad,erad,srad,pion,eion, &
                       sion,xnem,pele,eele,sele,pres,ener,entr,dpresdd, &
                       dpresdt,denerdd,denerdt,dentrdd,dentrdt,cv,cp, &
                       gam1,gam2,gam3,chit,chid,nabad,sound,etaele, &
                       detadt,detadd,xnefer,dxnedt,dxnedd,s

      double precision pgas,dpgasdd,dpgasdt,dpgasda,dpgasdz, &
                       egas,degasdd,degasdt,degasda,degasdz, &
                       sgas,dsgasdd,dsgasdt,dsgasda,dsgasdz, &
                       cv_gas,cp_gas,gam1_gas,gam2_gas,gam3_gas, &
                       chit_gas,chid_gas,nabad_gas,sound_gas


      double precision sioncon,forth,forpi,kergavo,ikavo,asoli3,light2
      parameter        (sioncon = (2.0d0 * pi * amu * kerg)/(h*h), &
                        forth   = 4.0d0/3.0d0, &
                        forpi   = 4.0d0 * pi, &
                        kergavo = kerg * avo, &
                        ikavo   = 1.0d0/kergavo, &
                        asoli3  = asol/3.0d0, &
                        light2  = clight * clight)

! for the abar derivatives
      double precision dpradda,deradda,dsradda, &
                       dpionda,deionda,dsionda, &
                       dpepda,deepda,dsepda, &
                       dpresda,denerda,dentrda, &
                       detada,dxneda

! for the zbar derivatives
      double precision dpraddz,deraddz,dsraddz, &
                       dpiondz,deiondz,dsiondz, &
                       dpepdz,deepdz,dsepdz, &
                       dpresdz,denerdz,dentrdz, &
                       detadz,dxnedz

! for the interpolations
      integer          iat,jat
      double precision free,df_d,df_t,df_dd,df_tt,df_dt
      double precision xt,xd,mxt,mxd, &
                       si0t,si1t,si2t,si0mt,si1mt,si2mt, &
                       si0d,si1d,si2d,si0md,si1md,si2md, &
                       dsi0t,dsi1t,dsi2t,dsi0mt,dsi1mt,dsi2mt, &
                       dsi0d,dsi1d,dsi2d,dsi0md,dsi1md,dsi2md, &
                       ddsi0t,ddsi1t,ddsi2t,ddsi0mt,ddsi1mt,ddsi2mt, &
                       ddsi0d,ddsi1d,ddsi2d,ddsi0md,ddsi1md,ddsi2md, &
                       z,psi0,dpsi0,ddpsi0,psi1,dpsi1,ddpsi1,psi2, &
                       dpsi2,ddpsi2,din,h5,fi(36), &
                       xpsi0,xdpsi0,xpsi1,xdpsi1,h3, &
                       w0t,w1t,w2t,w0mt,w1mt,w2mt, &
                       w0d,w1d,w2d,w0md,w1md,w2md


! for the uniform background coulomb correction
      double precision dsdd,dsda,lami,inv_lami,lamida,lamidd, &
                       plasg,plasgdd,plasgdt,plasgda,plasgdz, &
                       ecoul,decouldd,decouldt,decoulda,decouldz, &
                       pcoul,dpcouldd,dpcouldt,dpcoulda,dpcouldz, &
                       scoul,dscouldd,dscouldt,dscoulda,dscouldz, &
                       a1,b1,c1,d1,e1,a2,b2,c2,third,esqu
      parameter        (a1    = -0.898004d0, &
                        b1    =  0.96786d0, &
                        c1    =  0.220703d0, &
                        d1    = -0.86097d0, &
                        e1    =  2.5269d0, &
                        a2    =  0.29561d0, &
                        b2    =  1.9885d0, &
                        c2    =  0.288675d0, &
                        third =  1.0d0/3.0d0, &
                        esqu  =  qe * qe)


! quintic hermite polynomial statement functions
! psi0 and its derivatives
      psi0(z)   = z**3 * ( z * (-6.0d0*z + 15.0d0) -10.0d0) + 1.0d0
      dpsi0(z)  = z**2 * ( z * (-30.0d0*z + 60.0d0) - 30.0d0)
      ddpsi0(z) = z* ( z*( -120.0d0*z + 180.0d0) -60.0d0)


! psi1 and its derivatives
      psi1(z)   = z* ( z**2 * ( z * (-3.0d0*z + 8.0d0) - 6.0d0) + 1.0d0)
      dpsi1(z)  = z*z * ( z * (-15.0d0*z + 32.0d0) - 18.0d0) +1.0d0
      ddpsi1(z) = z * (z * (-60.0d0*z + 96.0d0) -36.0d0)


! psi2  and its derivatives
      psi2(z)   = 0.5d0*z*z*( z* ( z * (-z + 3.0d0) - 3.0d0) + 1.0d0)
      dpsi2(z)  = 0.5d0*z*( z*(z*(-5.0d0*z + 12.0d0) - 9.0d0) + 2.0d0)
      ddpsi2(z) = 0.5d0*(z*( z * (-20.0d0*z + 36.0d0) - 18.0d0) + 2.0d0)


! biquintic hermite polynomial statement function
      h5(i,j,w0t,w1t,w2t,w0mt,w1mt,w2mt,w0d,w1d,w2d,w0md,w1md,w2md)= &
             fi(1)  *w0d*w0t   + fi(2)  *w0md*w0t &
           + fi(3)  *w0d*w0mt  + fi(4)  *w0md*w0mt &
           + fi(5)  *w0d*w1t   + fi(6)  *w0md*w1t &
           + fi(7)  *w0d*w1mt  + fi(8)  *w0md*w1mt &
           + fi(9)  *w0d*w2t   + fi(10) *w0md*w2t &
           + fi(11) *w0d*w2mt  + fi(12) *w0md*w2mt &
           + fi(13) *w1d*w0t   + fi(14) *w1md*w0t &
           + fi(15) *w1d*w0mt  + fi(16) *w1md*w0mt &
           + fi(17) *w2d*w0t   + fi(18) *w2md*w0t &
           + fi(19) *w2d*w0mt  + fi(20) *w2md*w0mt &
           + fi(21) *w1d*w1t   + fi(22) *w1md*w1t &
           + fi(23) *w1d*w1mt  + fi(24) *w1md*w1mt &
           + fi(25) *w2d*w1t   + fi(26) *w2md*w1t &
           + fi(27) *w2d*w1mt  + fi(28) *w2md*w1mt &
           + fi(29) *w1d*w2t   + fi(30) *w1md*w2t &
           + fi(31) *w1d*w2mt  + fi(32) *w1md*w2mt &
           + fi(33) *w2d*w2t   + fi(34) *w2md*w2t &
           + fi(35) *w2d*w2mt  + fi(36) *w2md*w2mt



! cubic hermite polynomial statement functions
! psi0 & derivatives
      xpsi0(z)  = z * z * (2.0d0*z - 3.0d0) + 1.0
      xdpsi0(z) = z * (6.0d0*z - 6.0d0)


! psi1 & derivatives
      xpsi1(z)  = z * ( z * (z - 2.0d0) + 1.0d0)
      xdpsi1(z) = z * (3.0d0*z - 4.0d0) + 1.0d0


! bicubic hermite polynomial statement function
      h3(i,j,w0t,w1t,w0mt,w1mt,w0d,w1d,w0md,w1md) = &
             fi(1)  *w0d*w0t   +  fi(2)  *w0md*w0t &
           + fi(3)  *w0d*w0mt  +  fi(4)  *w0md*w0mt &
           + fi(5)  *w0d*w1t   +  fi(6)  *w0md*w1t &
           + fi(7)  *w0d*w1mt  +  fi(8)  *w0md*w1mt &
           + fi(9)  *w1d*w0t   +  fi(10) *w1md*w0t &
           + fi(11) *w1d*w0mt  +  fi(12) *w1md*w0mt &
           + fi(13) *w1d*w1t   +  fi(14) *w1md*w1t &
           + fi(15) *w1d*w1mt  +  fi(16) *w1md*w1mt



! popular format statements
01    format(1x,5(a,1pe11.3))
02    format(1x,a,1p4e16.8)
03    format(1x,4(a,1pe11.3))
04    format(1x,4(a,i4))



! start of pipeline loop, normal execution starts here
      eosfail = .false.
      do j=jlo_eos,jhi_eos

!       if (temp_row(j) .le. 0.0) stop 'temp less than 0 in helmeos'
!       if (den_row(j)  .le. 0.0) stop 'den less than 0 in helmeos'

       temp  = temp_row(j)
       den   = den_row(j)
       abar  = abar_row(j)
       zbar  = zbar_row(j)
       ytot1 = 1.0d0/abar
       ye    = max(1.0d-16,ytot1 * zbar)



! initialize
       deni    = 1.0d0/den
       tempi   = 1.0d0/temp
       kt      = kerg * temp
       ktinv   = 1.0d0/kt


! radiation section:
       prad    = asoli3 * temp * temp * temp * temp
       dpraddd = 0.0d0
       dpraddt = 4.0d0 * prad*tempi
       dpradda = 0.0d0
       dpraddz = 0.0d0

       erad    = 3.0d0 * prad*deni
       deraddd = -erad*deni
       deraddt = 3.0d0 * dpraddt*deni
       deradda = 0.0d0
       deraddz = 0.0d0

       srad    = (prad*deni + erad)*tempi
       dsraddd = (dpraddd*deni - prad*deni*deni + deraddd)*tempi
       dsraddt = (dpraddt*deni + deraddt - srad)*tempi
       dsradda = 0.0d0
       dsraddz = 0.0d0


! ion section:
        xni     = avo * ytot1 * den
        dxnidd  = avo * ytot1
        dxnida  = -xni * ytot1

        pion    = xni * kt
        dpiondd = dxnidd * kt
        dpiondt = xni * kerg
        dpionda = dxnida * kt
        dpiondz = 0.0d0

        eion    = 1.5d0 * pion*deni
        deiondd = (1.5d0 * dpiondd - eion)*deni
        deiondt = 1.5d0 * dpiondt*deni
        deionda = 1.5d0 * dpionda*deni
        deiondz = 0.0d0


! sackur-tetrode equation for the ion entropy of
! a single ideal gas characterized by abar
        x       = abar*abar*sqrt(abar) * deni/avo
        s       = sioncon * temp
        z       = x * s * sqrt(s)
        y       = log(z)

!        y       = 1.0d0/(abar*kt)
!        yy      = y * sqrt(y)
!        z       = xni * sifac * yy
!        etaion  = log(z)


        sion    = (pion*deni + eion)*tempi + kergavo * ytot1 * y
        dsiondd = (dpiondd*deni - pion*deni*deni + deiondd)*tempi &
                   - kergavo * deni * ytot1
        dsiondt = (dpiondt*deni + deiondt)*tempi - &
                  (pion*deni + eion) * tempi*tempi &
                  + 1.5d0 * kergavo * tempi*ytot1
        x       = avo*kerg/abar
        dsionda = (dpionda*deni + deionda)*tempi &
                  + kergavo*ytot1*ytot1* (2.5d0 - y)
        dsiondz = 0.0d0



! electron-positron section:


! assume complete ionization
        xnem    = xni * zbar


! enter the table with ye*den
        din = ye*den


! bomb proof the input
        if (temp .gt. t(jmax)) then
         write(6,01) 'temp=',temp,' t(jmax)=',t(jmax)
         write(6,*) 'temp too hot, off grid'
         write(6,*) 'setting eosfail to true and returning'
         eosfail = .true.
         return
        end if
        if (temp .lt. t(1)) then
         write(6,01) 'temp=',temp,' t(1)=',t(1)
         write(6,*) 'temp too cold, off grid'
         write(6,*) 'setting eosfail to true and returning'
         eosfail = .true.
         return
        end if
        if (din  .gt. d(imax)) then
         write(6,01) 'den*ye=',din,' d(imax)=',d(imax)
         write(6,*) 'ye*den too big, off grid'
         write(6,*) 'setting eosfail to true and returning'
         eosfail = .true.
         return
        end if
        if (din  .lt. d(1)) then
         write(6,01) 'ye*den=',din,' d(1)=',d(1)
         write(6,*) 'ye*den too small, off grid'
         write(6,*) 'setting eosfail to true and returning'
         eosfail = .true.
         return
        end if

! hash locate this temperature and density
        jat = int((log10(temp) - tlo)*tstpi) + 1
        jat = max(1,min(jat,jmax-1))
        iat = int((log10(din) - dlo)*dstpi) + 1
        iat = max(1,min(iat,imax-1))


! access the table locations only once
        fi(1)  = f(iat,jat)
        fi(2)  = f(iat+1,jat)
        fi(3)  = f(iat,jat+1)
        fi(4)  = f(iat+1,jat+1)
        fi(5)  = ft(iat,jat)
        fi(6)  = ft(iat+1,jat)
        fi(7)  = ft(iat,jat+1)
        fi(8)  = ft(iat+1,jat+1)
        fi(9)  = ftt(iat,jat)
        fi(10) = ftt(iat+1,jat)
        fi(11) = ftt(iat,jat+1)
        fi(12) = ftt(iat+1,jat+1)
        fi(13) = fd(iat,jat)
        fi(14) = fd(iat+1,jat)
        fi(15) = fd(iat,jat+1)
        fi(16) = fd(iat+1,jat+1)
        fi(17) = fdd(iat,jat)
        fi(18) = fdd(iat+1,jat)
        fi(19) = fdd(iat,jat+1)
        fi(20) = fdd(iat+1,jat+1)
        fi(21) = fdt(iat,jat)
        fi(22) = fdt(iat+1,jat)
        fi(23) = fdt(iat,jat+1)
        fi(24) = fdt(iat+1,jat+1)
        fi(25) = fddt(iat,jat)
        fi(26) = fddt(iat+1,jat)
        fi(27) = fddt(iat,jat+1)
        fi(28) = fddt(iat+1,jat+1)
        fi(29) = fdtt(iat,jat)
        fi(30) = fdtt(iat+1,jat)
        fi(31) = fdtt(iat,jat+1)
        fi(32) = fdtt(iat+1,jat+1)
        fi(33) = fddtt(iat,jat)
        fi(34) = fddtt(iat+1,jat)
        fi(35) = fddtt(iat,jat+1)
        fi(36) = fddtt(iat+1,jat+1)


! various differences
        xt  = max( (temp - t(jat))*dti_sav(jat), 0.0d0)
        xd  = max( (din - d(iat))*ddi_sav(iat), 0.0d0)
        mxt = 1.0d0 - xt
        mxd = 1.0d0 - xd

! the six density and six temperature basis functions
        si0t =   psi0(xt)
        si1t =   psi1(xt)*dt_sav(jat)
        si2t =   psi2(xt)*dt2_sav(jat)

        si0mt =  psi0(mxt)
        si1mt = -psi1(mxt)*dt_sav(jat)
        si2mt =  psi2(mxt)*dt2_sav(jat)

        si0d =   psi0(xd)
        si1d =   psi1(xd)*dd_sav(iat)
        si2d =   psi2(xd)*dd2_sav(iat)

        si0md =  psi0(mxd)
        si1md = -psi1(mxd)*dd_sav(iat)
        si2md =  psi2(mxd)*dd2_sav(iat)

! derivatives of the weight functions
        dsi0t =   dpsi0(xt)*dti_sav(jat)
        dsi1t =   dpsi1(xt)
        dsi2t =   dpsi2(xt)*dt_sav(jat)

        dsi0mt = -dpsi0(mxt)*dti_sav(jat)
        dsi1mt =  dpsi1(mxt)
        dsi2mt = -dpsi2(mxt)*dt_sav(jat)

        dsi0d =   dpsi0(xd)*ddi_sav(iat)
        dsi1d =   dpsi1(xd)
        dsi2d =   dpsi2(xd)*dd_sav(iat)

        dsi0md = -dpsi0(mxd)*ddi_sav(iat)
        dsi1md =  dpsi1(mxd)
        dsi2md = -dpsi2(mxd)*dd_sav(iat)

! second derivatives of the weight functions
        ddsi0t =   ddpsi0(xt)*dt2i_sav(jat)
        ddsi1t =   ddpsi1(xt)*dti_sav(jat)
        ddsi2t =   ddpsi2(xt)

        ddsi0mt =  ddpsi0(mxt)*dt2i_sav(jat)
        ddsi1mt = -ddpsi1(mxt)*dti_sav(jat)
        ddsi2mt =  ddpsi2(mxt)

!        ddsi0d =   ddpsi0(xd)*dd2i_sav(iat)
!        ddsi1d =   ddpsi1(xd)*ddi_sav(iat)
!        ddsi2d =   ddpsi2(xd)

!        ddsi0md =  ddpsi0(mxd)*dd2i_sav(iat)
!        ddsi1md = -ddpsi1(mxd)*ddi_sav(iat)
!        ddsi2md =  ddpsi2(mxd)


! the free energy
        free  = h5(iat,jat, &
                si0t,   si1t,   si2t,   si0mt,   si1mt,   si2mt, &
                si0d,   si1d,   si2d,   si0md,   si1md,   si2md)

! derivative with respect to density
        df_d  = h5(iat,jat, &
                si0t,   si1t,   si2t,   si0mt,   si1mt,   si2mt, &
                dsi0d,  dsi1d,  dsi2d,  dsi0md,  dsi1md,  dsi2md)


! derivative with respect to temperature
        df_t = h5(iat,jat, &
                dsi0t,  dsi1t,  dsi2t,  dsi0mt,  dsi1mt,  dsi2mt, &
                si0d,   si1d,   si2d,   si0md,   si1md,   si2md)

! derivative with respect to density**2
!        df_dd = h5(iat,jat,
!     1          si0t,   si1t,   si2t,   si0mt,   si1mt,   si2mt,
!     2          ddsi0d, ddsi1d, ddsi2d, ddsi0md, ddsi1md, ddsi2md)

! derivative with respect to temperature**2
        df_tt = h5(iat,jat, &
              ddsi0t, ddsi1t, ddsi2t, ddsi0mt, ddsi1mt, ddsi2mt, &
                si0d,   si1d,   si2d,   si0md,   si1md,   si2md)

! derivative with respect to temperature and density
        df_dt = h5(iat,jat, &
                dsi0t,  dsi1t,  dsi2t,  dsi0mt,  dsi1mt,  dsi2mt, &
                dsi0d,  dsi1d,  dsi2d,  dsi0md,  dsi1md,  dsi2md)



! now get the pressure derivative with density, chemical potential, and
! electron positron number densities
! get the interpolation weight functions
        si0t   =  xpsi0(xt)
        si1t   =  xpsi1(xt)*dt_sav(jat)

        si0mt  =  xpsi0(mxt)
        si1mt  =  -xpsi1(mxt)*dt_sav(jat)

        si0d   =  xpsi0(xd)
        si1d   =  xpsi1(xd)*dd_sav(iat)

        si0md  =  xpsi0(mxd)
        si1md  =  -xpsi1(mxd)*dd_sav(iat)


! derivatives of weight functions
        dsi0t  = xdpsi0(xt)*dti_sav(jat)
        dsi1t  = xdpsi1(xt)

        dsi0mt = -xdpsi0(mxt)*dti_sav(jat)
        dsi1mt = xdpsi1(mxt)

        dsi0d  = xdpsi0(xd)*ddi_sav(iat)
        dsi1d  = xdpsi1(xd)

        dsi0md = -xdpsi0(mxd)*ddi_sav(iat)
        dsi1md = xdpsi1(mxd)


! look in the pressure derivative only once
        fi(1)  = dpdf(iat,jat)
        fi(2)  = dpdf(iat+1,jat)
        fi(3)  = dpdf(iat,jat+1)
        fi(4)  = dpdf(iat+1,jat+1)
        fi(5)  = dpdft(iat,jat)
        fi(6)  = dpdft(iat+1,jat)
        fi(7)  = dpdft(iat,jat+1)
        fi(8)  = dpdft(iat+1,jat+1)
        fi(9)  = dpdfd(iat,jat)
        fi(10) = dpdfd(iat+1,jat)
        fi(11) = dpdfd(iat,jat+1)
        fi(12) = dpdfd(iat+1,jat+1)
        fi(13) = dpdfdt(iat,jat)
        fi(14) = dpdfdt(iat+1,jat)
        fi(15) = dpdfdt(iat,jat+1)
        fi(16) = dpdfdt(iat+1,jat+1)

! pressure derivative with density
        dpepdd  = h3(iat,jat, &
                       si0t,   si1t,   si0mt,   si1mt, &
                       si0d,   si1d,   si0md,   si1md)
        dpepdd  = max(ye * dpepdd,1.0d-30)



! look in the electron chemical potential table only once
        fi(1)  = ef(iat,jat)
        fi(2)  = ef(iat+1,jat)
        fi(3)  = ef(iat,jat+1)
        fi(4)  = ef(iat+1,jat+1)
        fi(5)  = eft(iat,jat)
        fi(6)  = eft(iat+1,jat)
        fi(7)  = eft(iat,jat+1)
        fi(8)  = eft(iat+1,jat+1)
        fi(9)  = efd(iat,jat)
        fi(10) = efd(iat+1,jat)
        fi(11) = efd(iat,jat+1)
        fi(12) = efd(iat+1,jat+1)
        fi(13) = efdt(iat,jat)
        fi(14) = efdt(iat+1,jat)
        fi(15) = efdt(iat,jat+1)
        fi(16) = efdt(iat+1,jat+1)


! electron chemical potential etaele
        etaele  = h3(iat,jat, &
                     si0t,   si1t,   si0mt,   si1mt, &
                     si0d,   si1d,   si0md,   si1md)


! derivative with respect to density
        x       = h3(iat,jat, &
                     si0t,   si1t,   si0mt,   si1mt, &
                    dsi0d,  dsi1d,  dsi0md,  dsi1md)
        detadd  = ye * x

! derivative with respect to temperature
        detadt  = h3(iat,jat, &
                    dsi0t,  dsi1t,  dsi0mt,  dsi1mt, &
                     si0d,   si1d,   si0md,   si1md)

! derivative with respect to abar and zbar
       detada = -x * din * ytot1
       detadz =  x * den * ytot1



! look in the number density table only once
        fi(1)  = xf(iat,jat)
        fi(2)  = xf(iat+1,jat)
        fi(3)  = xf(iat,jat+1)
        fi(4)  = xf(iat+1,jat+1)
        fi(5)  = xft(iat,jat)
        fi(6)  = xft(iat+1,jat)
        fi(7)  = xft(iat,jat+1)
        fi(8)  = xft(iat+1,jat+1)
        fi(9)  = xfd(iat,jat)
        fi(10) = xfd(iat+1,jat)
        fi(11) = xfd(iat,jat+1)
        fi(12) = xfd(iat+1,jat+1)
        fi(13) = xfdt(iat,jat)
        fi(14) = xfdt(iat+1,jat)
        fi(15) = xfdt(iat,jat+1)
        fi(16) = xfdt(iat+1,jat+1)

! electron + positron number densities
       xnefer   = h3(iat,jat, &
                     si0t,   si1t,   si0mt,   si1mt, &
                     si0d,   si1d,   si0md,   si1md)

! derivative with respect to density
       x        = h3(iat,jat, &
                     si0t,   si1t,   si0mt,   si1mt, &
                    dsi0d,  dsi1d,  dsi0md,  dsi1md)
       x = max(x,1.0d-30)
       dxnedd   = ye * x

! derivative with respect to temperature
       dxnedt   = h3(iat,jat, &
                    dsi0t,  dsi1t,  dsi0mt,  dsi1mt, &
                     si0d,   si1d,   si0md,   si1md)

! derivative with respect to abar and zbar
       dxneda = -x * din * ytot1
       dxnedz =  x  * den * ytot1


! the desired electron-positron thermodynamic quantities

! dpepdd at high temperatures and low densities is below the
! floating point limit of the subtraction of two large terms.
! since dpresdd doesn't enter the maxwell relations at all, use the
! bicubic interpolation done above instead of the formally correct expression
        x       = din * din
        pele    = x * df_d
        dpepdt  = x * df_dt
!        dpepdd  = ye * (x * df_dd + 2.0d0 * din * df_d)
        s       = dpepdd/ye - 2.0d0 * din * df_d
        dpepda  = -ytot1 * (2.0d0 * pele + s * din)
        dpepdz  = den*ytot1*(2.0d0 * din * df_d  +  s)


        x       = ye * ye
        sele    = -df_t * ye
        dsepdt  = -df_tt * ye
        dsepdd  = -df_dt * x
        dsepda  = ytot1 * (ye * df_dt * din - sele)
        dsepdz  = -ytot1 * (ye * df_dt * den  + df_t)


        eele    = ye*free + temp * sele
        deepdt  = temp * dsepdt
        deepdd  = x * df_d + temp * dsepdd
        deepda  = -ye * ytot1 * (free +  df_d * din) + temp * dsepda
        deepdz  = ytot1* (free + ye * df_d * den) + temp * dsepdz




! coulomb section:

! uniform background corrections only
! from yakovlev & shalybkov 1989
! lami is the average ion seperation
! plasg is the plasma coupling parameter

        z        = forth * pi
        s        = z * xni
        dsdd     = z * dxnidd
        dsda     = z * dxnida

        lami     = 1.0d0/s**third
        inv_lami = 1.0d0/lami
        z        = -third * lami
        lamidd   = z * dsdd/s
        lamida   = z * dsda/s

        plasg    = zbar*zbar*esqu*ktinv*inv_lami
        z        = -plasg * inv_lami
        plasgdd  = z * lamidd
        plasgda  = z * lamida
        plasgdt  = -plasg*ktinv * kerg
        plasgdz  = 2.0d0 * plasg/zbar


! yakovlev & shalybkov 1989 equations 82, 85, 86, 87
        if (plasg .ge. 1.0) then
         x        = plasg**(0.25d0)
         y        = avo * ytot1 * kerg
         ecoul    = y * temp * (a1*plasg + b1*x + c1/x + d1)
         pcoul    = third * den * ecoul
         scoul    = -y * (3.0d0*b1*x - 5.0d0*c1/x &
                    + d1 * (log(plasg) - 1.0d0) - e1)

         y        = avo*ytot1*kt*(a1 + 0.25d0/plasg*(b1*x - c1/x))
         decouldd = y * plasgdd
         decouldt = y * plasgdt + ecoul/temp
         decoulda = y * plasgda - ecoul/abar
         decouldz = y * plasgdz

         y        = third * den
         dpcouldd = third * ecoul + y*decouldd
         dpcouldt = y * decouldt
         dpcoulda = y * decoulda
         dpcouldz = y * decouldz


         y        = -avo*kerg/(abar*plasg)*(0.75d0*b1*x+1.25d0*c1/x+d1)
         dscouldd = y * plasgdd
         dscouldt = y * plasgdt
         dscoulda = y * plasgda - scoul/abar
         dscouldz = y * plasgdz


! yakovlev & shalybkov 1989 equations 102, 103, 104
        else if (plasg .lt. 1.0) then
         x        = plasg*sqrt(plasg)
         y        = plasg**b2
         z        = c2 * x - third * a2 * y
         pcoul    = -pion * z
         ecoul    = 3.0d0 * pcoul/den
         scoul    = -avo/abar*kerg*(c2*x -a2*(b2-1.0d0)/b2*y)

         s        = 1.5d0*c2*x/plasg - third*a2*b2*y/plasg
         dpcouldd = -dpiondd*z - pion*s*plasgdd
         dpcouldt = -dpiondt*z - pion*s*plasgdt
         dpcoulda = -dpionda*z - pion*s*plasgda
         dpcouldz = -dpiondz*z - pion*s*plasgdz

         s        = 3.0d0/den
         decouldd = s * dpcouldd - ecoul/den
         decouldt = s * dpcouldt
         decoulda = s * dpcoulda
         decouldz = s * dpcouldz

         s        = -avo*kerg/(abar*plasg)*(1.5d0*c2*x-a2*(b2-1.0d0)*y)
         dscouldd = s * plasgdd
         dscouldt = s * plasgdt
         dscoulda = s * plasgda - scoul/abar
         dscouldz = s * plasgdz
        end if


! bomb proof
        x   = prad + pion + pele + pcoul
        y   = erad + eion + eele + ecoul
        z   = srad + sion + sele + scoul

!        write(6,*) x,y,z
        if (x .le. 0.0 .or. y .le. 0.0 .or. z .le. 0.0) then
!        if (x .le. 0.0 .or. y .le. 0.0) then
!        if (x .le. 0.0) then

!         write(6,*)
!         write(6,*) 'coulomb corrections are causing a negative pressure'
!         write(6,*) 'setting all coulomb corrections to zero'
!         write(6,*)

         pcoul    = 0.0d0
         dpcouldd = 0.0d0
         dpcouldt = 0.0d0
         dpcoulda = 0.0d0
         dpcouldz = 0.0d0
         ecoul    = 0.0d0
         decouldd = 0.0d0
         decouldt = 0.0d0
         decoulda = 0.0d0
         decouldz = 0.0d0
         scoul    = 0.0d0
         dscouldd = 0.0d0
         dscouldt = 0.0d0
         dscoulda = 0.0d0
         dscouldz = 0.0d0
        end if


! sum all the gas components
       pgas    = pion + pele + pcoul
       egas    = eion + eele + ecoul
       sgas    = sion + sele + scoul

       dpgasdd = dpiondd + dpepdd + dpcouldd
       dpgasdt = dpiondt + dpepdt + dpcouldt
       dpgasda = dpionda + dpepda + dpcoulda
       dpgasdz = dpiondz + dpepdz + dpcouldz

       degasdd = deiondd + deepdd + decouldd
       degasdt = deiondt + deepdt + decouldt
       degasda = deionda + deepda + decoulda
       degasdz = deiondz + deepdz + decouldz

       dsgasdd = dsiondd + dsepdd + dscouldd
       dsgasdt = dsiondt + dsepdt + dscouldt
       dsgasda = dsionda + dsepda + dscoulda
       dsgasdz = dsiondz + dsepdz + dscouldz




! add in radiation to get the total
       pres    = prad + pgas
       ener    = erad + egas
       entr    = srad + sgas

       dpresdd = dpraddd + dpgasdd
       dpresdt = dpraddt + dpgasdt
       dpresda = dpradda + dpgasda
       dpresdz = dpraddz + dpgasdz

       denerdd = deraddd + degasdd
       denerdt = deraddt + degasdt
       denerda = deradda + degasda
       denerdz = deraddz + degasdz

       dentrdd = dsraddd + dsgasdd
       dentrdt = dsraddt + dsgasdt
       dentrda = dsradda + dsgasda
       dentrdz = dsraddz + dsgasdz


! for the gas
! the temperature and density exponents (c&g 9.81 9.82)
! the specific heat at constant volume (c&g 9.92)
! the third adiabatic exponent (c&g 9.93)
! the first adiabatic exponent (c&g 9.97)
! the second adiabatic exponent (c&g 9.105)
! the specific heat at constant pressure (c&g 9.98)
! and relativistic formula for the sound speed (c&g 14.29)

       zz        = pgas*deni
       zzi       = den/pgas
       chit_gas  = temp/pgas * dpgasdt
       chid_gas  = dpgasdd*zzi
       cv_gas    = degasdt
       x         = zz * chit_gas/(temp * cv_gas)
       gam3_gas  = x + 1.0d0
       gam1_gas  = chit_gas*x + chid_gas
       nabad_gas = x/gam1_gas
       gam2_gas  = 1.0d0/(1.0d0 - nabad_gas)
       cp_gas    = cv_gas * gam1_gas/chid_gas
       z         = 1.0d0 + (egas + light2)*zzi
       sound_gas = clight * sqrt(gam1_gas/z)



! for the totals
       zz    = pres*deni
       zzi   = den/pres
       chit  = temp/pres * dpresdt
       chid  = dpresdd*zzi
       cv    = denerdt
       x     = zz * chit/(temp * cv)
       gam3  = x + 1.0d0
       gam1  = chit*x + chid
       nabad = x/gam1
       gam2  = 1.0d0/(1.0d0 - nabad)
       cp    = cv * gam1/chid
       z     = 1.0d0 + (ener + light2)*zzi
       sound = clight * sqrt(gam1/z)



! maxwell relations; each is zero if the consistency is perfect
       x   = den * den

       dse = temp*dentrdt/denerdt - 1.0d0

       dpe = (denerdd*x + temp*dpresdt)/pres - 1.0d0

       dsp = -dentrdd*x/dpresdt - 1.0d0


! store this row
        ptot_row(j)   = pres
        dpt_row(j)    = dpresdt
        dpd_row(j)    = dpresdd
        dpa_row(j)    = dpresda
        dpz_row(j)    = dpresdz

        etot_row(j)   = ener
        det_row(j)    = denerdt
        ded_row(j)    = denerdd
        dea_row(j)    = denerda
        dez_row(j)    = denerdz

        stot_row(j)   = entr
        dst_row(j)    = dentrdt
        dsd_row(j)    = dentrdd
        dsa_row(j)    = dentrda
        dsz_row(j)    = dentrdz


        pgas_row(j)   = pgas
        dpgast_row(j) = dpgasdt
        dpgasd_row(j) = dpgasdd
        dpgasa_row(j) = dpgasda
        dpgasz_row(j) = dpgasdz

        egas_row(j)   = egas
        degast_row(j) = degasdt
        degasd_row(j) = degasdd
        degasa_row(j) = degasda
        degasz_row(j) = degasdz

        sgas_row(j)   = sgas
        dsgast_row(j) = dsgasdt
        dsgasd_row(j) = dsgasdd
        dsgasa_row(j) = dsgasda
        dsgasz_row(j) = dsgasdz


        prad_row(j)   = prad
        dpradt_row(j) = dpraddt
        dpradd_row(j) = dpraddd
        dprada_row(j) = dpradda
        dpradz_row(j) = dpraddz

        erad_row(j)   = erad
        deradt_row(j) = deraddt
        deradd_row(j) = deraddd
        derada_row(j) = deradda
        deradz_row(j) = deraddz

        srad_row(j)   = srad
        dsradt_row(j) = dsraddt
        dsradd_row(j) = dsraddd
        dsrada_row(j) = dsradda
        dsradz_row(j) = dsraddz


        pion_row(j)   = pion
        dpiont_row(j) = dpiondt
        dpiond_row(j) = dpiondd
        dpiona_row(j) = dpionda
        dpionz_row(j) = dpiondz

        eion_row(j)   = eion
        deiont_row(j) = deiondt
        deiond_row(j) = deiondd
        deiona_row(j) = deionda
        deionz_row(j) = deiondz

        sion_row(j)   = sion
        dsiont_row(j) = dsiondt
        dsiond_row(j) = dsiondd
        dsiona_row(j) = dsionda
        dsionz_row(j) = dsiondz

        xni_row(j)    = xni

        pele_row(j)   = pele
        ppos_row(j)   = 0.0d0
        dpept_row(j)  = dpepdt
        dpepd_row(j)  = dpepdd
        dpepa_row(j)  = dpepda
        dpepz_row(j)  = dpepdz

        eele_row(j)   = eele
        epos_row(j)   = 0.0d0
        deept_row(j)  = deepdt
        deepd_row(j)  = deepdd
        deepa_row(j)  = deepda
        deepz_row(j)  = deepdz

        sele_row(j)   = sele
        spos_row(j)   = 0.0d0
        dsept_row(j)  = dsepdt
        dsepd_row(j)  = dsepdd
        dsepa_row(j)  = dsepda
        dsepz_row(j)  = dsepdz

        xnem_row(j)   = xnem
        xne_row(j)    = xnefer
        dxnet_row(j)  = dxnedt
        dxned_row(j)  = dxnedd
        dxnea_row(j)  = dxneda
        dxnez_row(j)  = dxnedz
        xnp_row(j)    = 0.0d0
        zeff_row(j)   = zbar

        etaele_row(j) = etaele
        detat_row(j)  = detadt
        detad_row(j)  = detadd
        detaa_row(j)  = detada
        detaz_row(j)  = detadz
        etapos_row(j) = 0.0d0

        pcou_row(j)   = pcoul
        dpcout_row(j) = dpcouldt
        dpcoud_row(j) = dpcouldd
        dpcoua_row(j) = dpcoulda
        dpcouz_row(j) = dpcouldz

        ecou_row(j)   = ecoul
        decout_row(j) = decouldt
        decoud_row(j) = decouldd
        decoua_row(j) = decoulda
        decouz_row(j) = decouldz

        scou_row(j)   = scoul
        dscout_row(j) = dscouldt
        dscoud_row(j) = dscouldd
        dscoua_row(j) = dscoulda
        dscouz_row(j) = dscouldz

        plasg_row(j)  = plasg

        dse_row(j)    = dse
        dpe_row(j)    = dpe
        dsp_row(j)    = dsp

        cv_gas_row(j)    = cv_gas
        cp_gas_row(j)    = cp_gas
        gam1_gas_row(j)  = gam1_gas
        gam2_gas_row(j)  = gam2_gas
        gam3_gas_row(j)  = gam3_gas
        nabad_gas_row(j) = nabad_gas
        cs_gas_row(j)    = sound_gas

        cv_row(j)     = cv
        cp_row(j)     = cp
        gam1_row(j)   = gam1
        gam2_row(j)   = gam2
        gam3_row(j)   = gam3
        nabad_row(j)  = nabad
        cs_row(j)     = sound

! end of pipeline loop
      enddo
      return
      end





      subroutine invert_helm_pt_quiet
      include 'implno.dek'
      include 'const.dek'
      include 'vector_eos.dek'


! given the pressure, temperature, and composition
! find everything else

! it is assumed that ptot_row(j), temp_row(j), abar_row(j),
! zbar_row(j), and the pipe limits (jlo_eos:jhi_eos), have
! been set before calling this routine.

! on input den_row(j) conatins a guess for the density,
! on output den_row(j) contains the converged density.

! To get the greatest speed advantage, the eos should be fed a
! large pipe of data to work on.

! this version is quiet on all errors


! local variables
      integer          i,j,jlo_save,jhi_save
      double precision den,f,df,dennew,eostol,fpmin
      parameter        (eostol = 1.0d-8, &
                        fpmin  = 1.0d-14)


! initialize
      jlo_save = jlo_eos
      jhi_save = jhi_eos
      do j=jlo_eos, jhi_eos
       eoswrk01(j) = 0.0d0
       eoswrk02(j) = 0.0d0
       eoswrk03(j) = ptot_row(j)
       eoswrk04(j) = den_row(j)
      end do


! do the first newton loop with all elements in the pipe
      call helmeos

      do j = jlo_eos, jhi_eos

       f     = ptot_row(j)/eoswrk03(j) - 1.0d0
       df    = dpd_row(j)/eoswrk03(j)
       eoswrk02(j) = f/df

! limit excursions to factor of two changes
       den    = den_row(j)
       dennew = min(max(0.5d0*den,den - eoswrk02(j)),2.0d0*den)

! compute the error
       eoswrk01(j)  = abs((dennew - den)/den)

! store the new density, keep it within the table limits
       den_row(j)  = min(1.0d14,max(dennew,1.0d-11))

      enddo



! now loop over each element of the pipe individually
      do j = jlo_save, jhi_save

       do i=2,40

        if (eoswrk01(j) .lt. eostol .or. &
            abs(eoswrk02(j)) .le. fpmin) goto 20

        jlo_eos = j
        jhi_eos = j

        call helmeos

        f     = ptot_row(j)/eoswrk03(j) - 1.0d0
        df    = dpd_row(j)/eoswrk03(j)
        eoswrk02(j) = f/df

! limit excursions to factor of two changes
        den    = den_row(j)
        dennew = min(max(0.5d0*den,den - eoswrk02(j)),2.0d0*den)

! compute the error
        eoswrk01(j)  = abs((dennew - den)/den)

! store the new density, keep it within the table limits
        den_row(j)  = min(1.0d14,max(dennew,1.0d-11))

! end of netwon loop
       end do


! we did not converge if we land here
!      write(6,*)
!      write(6,*) 'newton-raphson failed in routine invert_helm_pt'
!      write(6,*) 'pipeline element',j
!      write(6,01) 'pres  =',eoswrk03(j)
! 01   format(1x,5(a,1pe16.8))
!      write(6,01) 'error =',eoswrk01(j),
!     1            '  eostol=',eostol,'  fpmin =',fpmin
!      write(6,01) 'den   =',den_row(j),'  denold=',eoswrk04(j)
!      write(6,01) 'f/df  =',eoswrk02(j),' f   =',f,    ' df    =',df
!      write(6,*)
!      stop 'could not find a density in routine invert_helm_pt'



! land here if newton loop converged, back for another pipe element
 20    continue
      end do



! call eos one more time with the converged value of the density

      jlo_eos = jlo_save
      jhi_eos = jhi_save

      call helmeos

      return
      end subroutine invert_helm_pt_quiet
